""".. Ignore pydocstyle D400.

===========================
Elastic Management commands
===========================

Elastic app includes following Django management commands:

.. automodule:: resolwe.elastic.management.commands.elastic_index

.. automodule:: resolwe.elastic.management.commands.elastic_mapping

.. automodule:: resolwe.elastic.management.commands.elastic_purge

"""
