# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-09-05 09:28
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0017_refactor_relations_3'),
    ]

    operations = [
        migrations.AddField(
            model_name='process',
            name='is_active',
            field=models.BooleanField(default=True, verbose_name='active'),
        ),
    ]
