# -*- coding: utf-8 -*-
# Generated by Django 1.11.14 on 2018-09-21 11:13
from __future__ import unicode_literals

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('flow', '0020_process_category_default'),
    ]

    operations = [
        migrations.AddField(
            model_name='collection',
            name='tags',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=255), default=list, size=None),
        ),
    ]
